package com.yonge.netty.server.handler.message;

import io.netty.buffer.ByteBuf;
import io.netty.buffer.ByteBufUtil;
import io.netty.channel.Channel;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.handler.codec.http.websocketx.BinaryWebSocketFrame;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Collectors;

import javax.sound.sampled.AudioFormat;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;

import com.yonge.audio.analysis.AudioFloatConverter;
import com.yonge.audio.utils.ArrayUtil;
import com.yonge.nettty.dto.UserChannelContext;
import com.yonge.nettty.dto.WebSocketResponse;
import com.yonge.nettty.entity.MusicXmlBasicInfo;
import com.yonge.netty.server.handler.ChannelContextConstants;
import com.yonge.netty.server.handler.NettyChannelManager;
import com.yonge.netty.server.processor.WaveformWriter;
import com.yonge.netty.server.service.UserChannelContextService;

@Component
@ChannelHandler.Sharable
public class BinaryWebSocketFrameHandler extends SimpleChannelInboundHandler<BinaryWebSocketFrame> implements ApplicationContextAware,InitializingBean {

	private final static Logger LOGGER = LoggerFactory.getLogger(BinaryWebSocketFrameHandler.class);
	
	@Autowired
	private NettyChannelManager nettyChannelManager;

	@Autowired
	private UserChannelContextService userChannelContextService;
	
	private ApplicationContext applicationContext;
	
	private Map<String, BinaryMessageHandler> handlerMap;

	/**
	 * @describe 采样率
	 */
	private float sampleRate = 44100;

	/**
	 * 每个采样大小（Bit）
	 */
	private int bitsPerSample = 16;

	/**
	 * 通道数
	 */
	private int channels = 1;

	/**
	 * @describe 采样大小
	 */
	private int bufferSize = 1024 * 4;

	private boolean signed = true;

	private boolean bigEndian = false;

	private AudioFormat audioFormat = new AudioFormat(sampleRate, bitsPerSample, channels, signed, bigEndian);

	private AudioFloatConverter converter = AudioFloatConverter.getConverter(audioFormat);

	private String tmpFileDir = "/mdata/soundCompare/";
	
	private SimpleDateFormat sdf =new SimpleDateFormat("yyMMddHHmmSS");
	
	@Override
	protected void channelRead0(ChannelHandlerContext ctx, BinaryWebSocketFrame frame) throws Exception {

		Channel channel = ctx.channel();

		ByteBuf buf = frame.content().retain();

		try {
			byte[] datas = ByteBufUtil.getBytes(buf);

			String user = nettyChannelManager.getUser(channel);
			
			String action = channel.attr(ChannelContextConstants.CHANNEL_ATTR_KEY_ACTION).get();
			
			if(handlerMap == null){
				LOGGER.error("消息处理器没有初始化");
			}
			BinaryMessageHandler handler = handlerMap.get(action);
			
			switch (action) {
			case "PITCH_DETECTION":
				handler.handler(user, channel, datas);
				break;
			case "SOUND_COMPARE":
				
				UserChannelContext channelContext = userChannelContextService.getChannelContext(channel);

				if (channelContext == null) {
					return;
				}
				
				// 写录音文件
				WaveformWriter waveFileProcessor = channelContext.getWaveFileProcessor();
				if (waveFileProcessor == null) {
					File file = new File(tmpFileDir + user + "_" + sdf.format(new Date()) + ".wav");
					waveFileProcessor = new WaveformWriter(file.getAbsolutePath());
					channelContext.setWaveFileProcessor(waveFileProcessor);
				}
				waveFileProcessor.process(datas);
				
				datas = channelContext.skipHeader(datas);

				if (datas.length == 0) {
					return;
				}
				
				channelContext.setChannelBufferBytes(ArrayUtil.mergeByte(channelContext.getChannelBufferBytes(), datas));
				
				int totalLength = channelContext.getChannelBufferBytes().length;
				
				while (totalLength >= bufferSize) {
					byte[] bufferData = ArrayUtil.extractByte(channelContext.getChannelBufferBytes(), 0, bufferSize - 1);

					if (bufferSize != totalLength) {
						channelContext.setChannelBufferBytes(ArrayUtil.extractByte(channelContext.getChannelBufferBytes(), bufferSize, totalLength - 1));
					} else {
						channelContext.setChannelBufferBytes(new byte[0]);
					}

					float[] sampleFloats = new float[bufferSize / 2];

					converter.toFloatArray(bufferData, sampleFloats);

					channelContext.handle(sampleFloats, audioFormat);

					MusicXmlBasicInfo musicXmlBasicInfo = channelContext.getMusicXmlBasicInfo(null);
					int sectionIndex = channelContext.getEvaluatingSectionIndex().get();

					// 评分
					int score = channelContext.evaluateForSection(sectionIndex, musicXmlBasicInfo.getSubjectId());
					if (score >= 0) {

						Map<String, Object> params = new HashMap<String, Object>();
						params.put("score", score);
						params.put("measureIndex", sectionIndex);

						WebSocketResponse<Map<String, Object>> resp = new WebSocketResponse<Map<String, Object>>("measureScore", params);

						nettyChannelManager.sendTextMessage(user, resp);
					}

					totalLength = channelContext.getChannelBufferBytes().length;
				}
				break;

			default:
				break;
			}

		} finally {
			buf.release();
		}
	}

	@Override
	public void afterPropertiesSet() throws Exception {
		handlerMap = applicationContext.getBeansOfType(BinaryMessageHandler.class).values().stream()
				.collect(Collectors.toMap(BinaryMessageHandler::getAction, t -> t));
	}

	@Override
	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
		this.applicationContext = applicationContext;
	}

}
